from pymimir import GlobalFaithfulAbstraction

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


def test_global_faithful_abstraction():
    """ Test the parallel construction of parallel faithful abstraction
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath_1 = str(ROOT_DIR / "data" / "gripper" / "p-1-0.pddl")
    problem_filepath_2 = str(ROOT_DIR / "data" / "gripper" / "p-1-0.pddl")
    problem_filepath_3 = str(ROOT_DIR / "data" / "gripper" / "p-2-0.pddl")
    problem_filepaths = [problem_filepath_1, problem_filepath_2, problem_filepath_3]

    abstractions = GlobalFaithfulAbstraction.create(domain_filepath, problem_filepaths)

    # Problem 1 was pruned because it has 0 global non isomorphic states.
    assert len(abstractions) == 2

    # Reduced from 8 to 6 abstract states.
    abstraction_0 = abstractions[0]
    assert abstraction_0.get_num_states() == 6
    assert abstraction_0.get_num_isomorphic_states() == 0
    assert abstraction_0.get_num_non_isomorphic_states() == 6
    # Reduced from 28 to 12 abstract states.
    abstraction_1 = abstractions[1]
    assert abstraction_1.get_num_states() == 12
    assert abstraction_1.get_num_isomorphic_states() == 0
    assert abstraction_1.get_num_non_isomorphic_states() == 12



from pymimir import PDDLParser, LiftedApplicableActionGenerator, BrFSAlgorithm, SearchStatus

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


def test_breadth_first_search():
    """ Test parsing a PDDL domain and problem file.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    parser = PDDLParser(domain_filepath, problem_filepath)
    lifted_applicable_action_generator = LiftedApplicableActionGenerator(parser.get_problem(), parser.get_pddl_factories())
    breadth_first_search_algorithm = BrFSAlgorithm(lifted_applicable_action_generator)
    search_status, plan = breadth_first_search_algorithm.find_solution()

    assert search_status == SearchStatus.SOLVED
    assert len(plan) == 3



# Import all classes for better IDE support

# Formalism
from _pymimir import (
    Action,
    ActionList,
    AssignOperatorEnum,
    Axiom,
    AxiomList,
    BinaryOperatorEnum,
    DerivedAtom,
    DerivedAtomList,
    DerivedGroundAtom,
    DerivedGroundAtomList,
    DerivedGroundLiteral,
    DerivedGroundLiteralList,
    DerivedLiteral,
    DerivedLiteralList,
    DerivedPredicate,
    DerivedPredicateList,
    Domain,
    DomainList,
    EffectConditional,
    EffectConditionalList,
    EffectSimple,
    EffectSimpleList,
    EffectUniversal,
    EffectUniversalList,
    FluentAtom,
    FluentAtomList,
    FluentGroundAtom,
    FluentGroundAtomList,
    FluentGroundLiteral,
    FluentGroundLiteralList,
    FluentLiteral,
    FluentLiteralList,
    FluentPredicate,
    FluentPredicateList,
    Function,
    FunctionList,
    FunctionExpression,
    FunctionExpressionVariantList,
    FunctionExpressionBinaryOperator,
    FunctionExpressionFunction,
    FunctionExpressionMinus,
    FunctionExpressionMultiOperator,
    FunctionExpressionNumber,
    FunctionSkeleton,
    FunctionSkeletonList,
    GroundFunction,
    GroundFunctionList,
    GroundFunctionExpression,
    GroundFunctionExpressionVariantList,
    GroundFunctionExpressionBinaryOperator,
    GroundFunctionExpressionFunction,
    GroundFunctionExpressionMinus,
    GroundFunctionExpressionMultiOperator,
    GroundFunctionExpressionNumber,
    MultiOperatorEnum,
    NumericFluent,
    NumericFluentList,
    Object,
    ObjectList,
    OptimizationMetric,
    OptimizationMetricEnum,
    PDDLFactories,
    PDDLParser,
    Problem,
    ProblemList,
    Requirements,
    RequirementEnum,
    StaticAtom,
    StaticAtomList,
    StaticGroundAtom,
    StaticGroundAtomList,
    StaticGroundLiteral,
    StaticGroundLiteralList,
    StaticLiteral,
    StaticLiteralList,
    StaticPredicate,
    StaticPredicateList,
    Term,
    TermVariantList,
    TermObject,
    TermVariable,
    Variable,
    VariableList
)

# Search
from _pymimir import (
    AStarAlgorithm,
    AStarAlgorithmEventHandlerBase,
    AStarAlgorithmStatistics,
    BlindHeuristic,
    BrFSAlgorithm,
    BrFSAlgorithmStatistics,
    ConditionalEffect,
    DebugAStarAlgorithmEventHandler,
    DebugBrFSAlgorithmEventHandler,
    DebugGroundedApplicableActionGeneratorEventHandler,
    DebugLiftedApplicableActionGeneratorEventHandler,
    DefaultBrFSAlgorithmEventHandler,
    DefaultAStarAlgorithmEventHandler,
    DefaultGroundedApplicableActionGeneratorEventHandler,
    DefaultIWAlgorithmEventHandler,
    DefaultLiftedApplicableActionGeneratorEventHandler,
    DefaultSIWAlgorithmEventHandler,
    FlatSimpleEffect,
    FluentAndDerivedMapper,
    GroundAction,
    GroundActionList,
    GroundActionSpan,
    GroundedApplicableActionGenerator,
    IApplicableActionGenerator,
    IAlgorithm,
    IAStarAlgorithmEventHandler,
    IBrFSAlgorithmEventHandler,
    IIWAlgorithmEventHandler,
    IGroundedApplicableActionGeneratorEventHandler,
    IHeuristic,
    ILiftedApplicableActionGeneratorEventHandler,
    ISIWAlgorithmEventHandler,
    IWAlgorithm,
    IWAlgorithmStatistics,
    LiftedApplicableActionGenerator,
    SearchNodeStatus,
    SearchStatus,
    SIWAlgorithm,
    SIWAlgorithmStatistics,
    State,
    StateList,
    StateRepository,
    StateSpan,
    StateIndexGroupedVector,
    StripsActionEffect,
    StripsActionPrecondition,
    TupleIndexMapper
)

# Dataset
from _pymimir import (
    Abstraction,
    StateVertex,
    GroundActionEdge,
    GroundActionsEdge,
    FaithfulAbstraction,
    FaithfulAbstractionOptions,
    FaithfulAbstractionsOptions,
    FaithfulAbstractStateVertex,
    GlobalFaithfulAbstractState,
    GlobalFaithfulAbstraction,
    StateSpace,
    StateSpaceOptions,
    StateSpacesOptions
)

# Graphs (classes)
from _pymimir import (
    Certificate,
    ColoredVertex,
    ColorFunction,
    DenseNautyGraph,
    EmptyVertex,
    EmptyEdge,
    ObjectGraphPruningStrategy,
    ObjectGraphPruningStrategyEnum,
    ProblemColorFunction,
    StaticDigraph,
    StaticVertexColoredDigraph,
    SparseNautyGraph,
    TupleGraphVertex,
    TupleGraphVertexSpan,
    TupleGraphVertexIndexGroupedVector,
    TupleGraph,
    TupleGraphFactory
)

# Graphs (free functions)
from _pymimir import (
    compute_vertex_colors,
    compute_sorted_vertex_colors,
    create_object_graph
)



from pymimir import PDDLParser, PDDLFactories, Problem, LiftedApplicableActionGenerator, State, StateRepository, GroundAction, AStarAlgorithm, SearchStatus, IHeuristic, DefaultAStarAlgorithmEventHandler, AStarAlgorithmEventHandlerBase

from pathlib import Path
from typing import List

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


class CustomBlindHeuristic(IHeuristic):
    def __init__(self):
        IHeuristic.__init__(self)  # Without this, a TypeError is raised.

    def compute_heuristic(self, state : State) -> float:
        return 0.


class CustomGoalCountHeuristic(IHeuristic):
    def __init__(self, problem : Problem, pddl_factories: PDDLFactories):
        IHeuristic.__init__(self)  # Without this, a TypeError is raised.
        self.problem = problem
        self.pddl_factories = pddl_factories

        self.num_goal_literals = len(self.problem.get_fluent_goal_condition()) + len(self.problem.get_derived_goal_condition())

    def compute_heuristic(self, state : State) -> float:
        num_satisfied_goal_literals = 0
        for literal in self.problem.get_fluent_goal_condition():
            if state.literal_holds(literal):
                num_satisfied_goal_literals += 1
        for literal in self.problem.get_derived_goal_condition():
            if state.literal_holds(literal):
                num_satisfied_goal_literals += 1

        return self.num_goal_literals - num_satisfied_goal_literals


class CustomAStarAlgorithmEventHandler(AStarAlgorithmEventHandlerBase):
    def __init__(self, quiet = True):
        """
        AStarAlgorithmEventHandlerBase collects statistics and
        calls callback implementations only if quiet = False.
        """
        AStarAlgorithmEventHandlerBase.__init__(self, quiet)  # Without this, a TypeError is raised.

    def on_expand_state_impl(self, state : State, problem : Problem, pddl_factories : PDDLFactories):
        pass

    def on_generate_state_impl(self, state : State, action : GroundAction, problem : Problem, pddl_factories : PDDLFactories):
        pass

    def on_generate_state_relaxed_impl(self, state : State, action : GroundAction, problem : Problem, pddl_factories : PDDLFactories):
        pass

    def on_generate_state_not_relaxed_impl(self, state : State, action : GroundAction, problem : Problem, pddl_factories : PDDLFactories):
        pass

    def on_close_state_impl(self, state : State, problem : Problem, pddl_factories : PDDLFactories):
        pass

    def on_finish_f_layer_impl(self, f_value: float, num_expanded_states : int, num_generated_states : int):
        pass

    def on_prune_state_impl(self, state : State, problem : Problem, pddl_factories : PDDLFactories):
        pass

    def on_start_search_impl(self, start_state : State, problem : Problem, pddl_factories : PDDLFactories):
        pass

    def on_end_search_impl(self):
        pass

    def on_solved_impl(self, ground_action_plan: List[GroundAction]):
        pass

    def on_unsolvable_impl(self):
        pass

    def on_exhausted_impl(self):
        pass

def test_astar_search():
    """ Test parsing a PDDL domain and problem file.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    parser = PDDLParser(domain_filepath, problem_filepath)
    lifted_applicable_action_generator = LiftedApplicableActionGenerator(parser.get_problem(), parser.get_pddl_factories())
    state_repository = StateRepository(lifted_applicable_action_generator)

    blind_heuristic = CustomBlindHeuristic()
    goal_count_heuristic = CustomGoalCountHeuristic(parser.get_problem(), parser.get_pddl_factories())

    event_handler = CustomAStarAlgorithmEventHandler(False)
    astar_search_algorithm = AStarAlgorithm(lifted_applicable_action_generator, state_repository, goal_count_heuristic, event_handler)
    search_status, plan = astar_search_algorithm.find_solution()

    assert search_status == SearchStatus.SOLVED
    assert len(plan) == 3



from pymimir import PDDLParser, LiftedApplicableActionGenerator, StateRepository

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


def test_state_repository_ownership():
    """ Test correct implementation of ownership semantics of `StateRepository`.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    pddl_parser = PDDLParser(domain_filepath, problem_filepath)

    state_repository = StateRepository(LiftedApplicableActionGenerator(pddl_parser.get_problem(), pddl_parser.get_pddl_factories()))
    initial_state = state_repository.get_or_create_initial_state()

    del state_repository

    assert initial_state.to_string(pddl_parser.get_problem(), pddl_parser.get_pddl_factories())



from pymimir import PDDLParser, LiftedApplicableActionGenerator, StateRepository

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent).absolute()


def test_applicable_action_generator_ownership():
    """ Test correct implementation of ownership semantics of `ApplicableActionGenerator`.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    pddl_parser = PDDLParser(domain_filepath, problem_filepath)

    applicable_action_generator = LiftedApplicableActionGenerator(pddl_parser.get_problem(), pddl_parser.get_pddl_factories())
    state_repository = StateRepository(applicable_action_generator)
    initial_state = state_repository.get_or_create_initial_state()
    actions = applicable_action_generator.compute_applicable_actions(initial_state)

    del state_repository
    del applicable_action_generator

    assert actions
    assert actions[0].to_string(pddl_parser.get_pddl_factories())
